#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Optional

import torch
from torch import nn

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import AUDIO, SEQUENCE, TEXT, TIMESERIES
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder, register_sequence_encoder
from ludwig.modules.attention_modules import TransformerStack
from ludwig.modules.convolutional_modules import Conv1DStack, ParallelConv1D, ParallelConv1DStack
from ludwig.modules.embedding_modules import EmbedSequence, TokenAndPositionEmbedding
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.modules.recurrent_modules import RecurrentStack
from ludwig.modules.reduction_modules import SequenceReducer
from ludwig.schema.encoders.sequence_encoders import (
    ParallelCNNConfig,
    SequenceEmbedConfig,
    SequencePassthroughConfig,
    StackedCNNConfig,
    StackedCNNRNNConfig,
    StackedParallelCNNConfig,
    StackedRNNConfig,
    StackedTransformerConfig,
)

logger = logging.getLogger(__name__)


class SequenceEncoder(Encoder):
    pass


@DeveloperAPI
@register_encoder("passthrough", [SEQUENCE, TEXT, TIMESERIES])
class SequencePassthroughEncoder(SequenceEncoder):
    def __init__(
        self,
        reduce_output: str = None,
        max_sequence_length: int = 256,
        encoding_size: int = None,
        encoder_config=None,
        **kwargs,
    ):
        """
        :param reduce_output: defines how to reduce the output tensor along
               the `s` sequence length dimension if the rank of the tensor
               is greater than 2. Available values are: `sum`,
               `mean` or `avg`, `max`, `concat` (concatenates along
               the first dimension), `last` (returns the last vector of the
               first dimension) and `None` or `null` (which does not reduce
               and returns the full tensor).
        :param max_sequence_length: The maximum sequence length.
        :param encoding_size: The size of the encoding vector, or None if sequence elements are scalars.
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output, max_sequence_length=max_sequence_length, encoding_size=encoding_size
        )
        if self.reduce_output is None:
            self.supports_masking = True

    def forward(self, input_sequence, mask=None):
        """
        :param input_sequence: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32 or
                      [batch x sequence length x encoding size], type torch.float32
        :type input_sequence: Tensor
        :param mask: Sequence mask (not yet implemented).
               Shape: [batch x sequence length]
        :type mask: Tensor
        """
        input_sequence = input_sequence.type(torch.float32)
        while len(input_sequence.shape) < 3:
            input_sequence = input_sequence.unsqueeze(-1)
        hidden = self.reduce_sequence(input_sequence)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return SequencePassthroughConfig


@DeveloperAPI
@register_encoder("embed", [SEQUENCE, TEXT])
class SequenceEmbedEncoder(SequenceEncoder):
    def __init__(
        self,
        vocab,
        max_sequence_length,
        representation="dense",
        embedding_size=256,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        weights_initializer=None,
        dropout=0,
        reduce_output="sum",
        encoder_config=None,
        **kwargs,
    ):
        """
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param max_sequence_length: The maximum sequence length.
        :type max_sequence_length: int
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memory and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :type embeddings_on_cpu: Boolean
        :param weights_initializer: the initializer to use. If `None`, the default
               initialized of each variable is used (`xavier_uniform`
               in most cases). Options are: `constant`, `identity`, `zeros`,
                `ones`, `orthogonal`, `normal`, `uniform`,
                `truncated_normal`, `variance_scaling`, `xavier_normal`,
                `xavier_uniform`, `xavier_normal`,
                `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                Alternatively it is possible to specify a dictionary with
                a key `type` that identifies the type of initializer and
                other keys for its parameters, e.g.
                `{type: normal, mean: 0, stddev: 0}`.
                To know the parameters of each initializer, please refer to
                PyTorch's documentation.
        :type weights_initializer: str
        :param dropout: Tensor (torch.float) The dropout probability.
        :type dropout: Tensor
        :param reduce_output: defines how to reduce the output tensor along
               the `s` sequence length dimension if the rank of the tensor
               is greater than 2. Available values are: `sum`,
               `mean` or `avg`, `max`, `concat` (concatenates along
               the first dimension), `last` (returns the last vector of the
               first dimension) and `None` or `null` (which does not reduce
               and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")
        self.embedding_size = embedding_size
        self.max_sequence_length = max_sequence_length

        self.reduce_output = reduce_output
        if self.reduce_output is None:
            self.supports_masking = True

        logger.debug("  EmbedSequence")
        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            max_sequence_length=max_sequence_length,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=max_sequence_length,
            encoding_size=self.embed_sequence.output_shape[-1],
        )

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented in EmbedSequence)
        """
        embedded_sequence = self.embed_sequence(inputs, mask=mask)
        hidden = self.reduce_sequence(embedded_sequence)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return SequenceEmbedConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        return self.reduce_sequence.output_shape


@DeveloperAPI
@register_sequence_encoder("parallel_cnn")
@register_encoder("parallel_cnn", [AUDIO, SEQUENCE, TEXT, TIMESERIES])
class ParallelCNN(SequenceEncoder):
    def __init__(
        self,
        should_embed=True,
        vocab=None,
        representation="dense",
        embedding_size=256,
        max_sequence_length=None,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        conv_layers=None,
        num_conv_layers=None,
        filter_size=3,
        num_filters=256,
        pool_function="max",
        pool_size=None,
        fc_layers=None,
        num_fc_layers=None,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        activation="relu",
        dropout=0,
        reduce_output="max",
        encoder_config=None,
        **kwargs,
    ):
        # todo: revise docstring
        """
        :param should_embed: If True the input sequence is expected
               to be made of integers and will be mapped into embeddings
        :type should_embed: Boolean
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: Str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memroy and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param conv_layers: it is a list of dictionaries containing
               the parameters of all the convolutional layers. The length
               of the list determines the number of parallel convolutional
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `filter_size`, `num_filters`, `pool`,
               `norm`, and `activation`. If any of those values
               is missing from the dictionary, the default one specified
               as a parameter of the encoder will be used instead. If both
               `conv_layers` and `num_conv_layers` are `None`, a default
               list will be assigned to `conv_layers` with the value
               `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
               {filter_size: 5}]`.
        :type conv_layers: List
        :param num_conv_layers: if `conv_layers` is `None`, this is
               the number of parallel convolutional layers.
        :type num_conv_layers: Integer
        :param filter_size:  if a `filter_size` is not already specified in
               `conv_layers` this is the default `filter_size` that
               will be used for each layer. It indicates how wide is
               the 1d convolutional filter.
        :type filter_size: Integer
        :param num_filters: if a `num_filters` is not already specified in
               `conv_layers` this is the default `num_filters` that
               will be used for each layer. It indicates the number
               of filters, and by consequence the output channels of
               the 1d convolution.
        :type num_filters: Integer
        :param pool_size: if a `pool_size` is not already specified
              in `conv_layers` this is the default `pool_size` that
              will be used for each layer. It indicates the size of
              the max pooling that will be performed along the `s` sequence
              dimension after the convolution operation.
        :type pool_size: Integer
        :param fc_layers: it is a list of dictionaries containing
               the parameters of all the fully connected layers. The length
               of the list determines the number of stacked fully connected
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `output_size`, `norm` and `activation`.
               If any of those values is missing from
               the dictionary, the default one specified as a parameter of
               the encoder will be used instead. If both `fc_layers` and
               `num_fc_layers` are `None`, a default list will be assigned
               to `fc_layers` with the value
               `[{output_size: 512}, {output_size: 256}]`
               (only applies if `reduce_output` is not `None`).
        :type fc_layers: List
        :param num_fc_layers: if `fc_layers` is `None`, this is the number
               of stacked fully connected layers (only applies if
               `reduce_output` is not `None`).
        :type num_fc_layers: Integer
        :param output_size: if a `output_size` is not already specified in
               `fc_layers` this is the default `output_size` that will be used
               for each layer. It indicates the size of the output
               of a fully connected layer.
        :type output_size: Integer
        :param norm: if a `norm` is not already specified in `conv_layers`
               or `fc_layers` this is the default `norm` that will be used
               for each layer. It indicates the norm of the output.
        :type norm: str
        :param activation: Default activation function to use
        :type activation: Str
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        :param initializer: the initializer to use. If `None` it uses
               `xavier_uniform`. Options are: `constant`, `identity`,
               `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
               `truncated_normal`, `variance_scaling`, `xavier_normal`,
               `xavier_uniform`, `xavier_normal`,
               `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
               Alternatively it is possible to specify a dictionary with
               a key `type` that identifies the type of initializer and
               other keys for its parameters,
               e.g. `{type: normal, mean: 0, stddev: 0}`.
               To know the parameters of each initializer, please refer
               to PyTorch's documentation.
        :type initializer: str
        :param reduce_output: defines how to reduce the output tensor of
               the convolutional layers along the `s` sequence length
               dimension if the rank of the tensor is greater than 2.
               Available values are: `sum`, `mean` or `avg`, `max`, `concat`
               (concatenates along the first dimension), `last` (returns
               the last vector of the first dimension) and `None` or `null`
               (which does not reduce and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.max_sequence_length = max_sequence_length

        if conv_layers is not None and num_conv_layers is None:
            # use custom-defined layers
            self.conv_layers = conv_layers
            self.num_conv_layers = len(conv_layers)
        elif conv_layers is None and num_conv_layers is not None:
            # generate num_conv_layers with default parameters
            self.conv_layers = None
            self.num_conv_layers = num_conv_layers
        elif conv_layers is None and num_conv_layers is None:
            # use default layers with varying filter sizes
            self.conv_layers = [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}]
            self.num_conv_layers = 4
        else:
            raise ValueError("Invalid layer parametrization, use either conv_layers or num_conv_layers")

        # The user is expected to provide fc_layers or num_fc_layers
        # The following logic handles the case where the user either provides
        # both or neither.
        if fc_layers is None and num_fc_layers is None:
            # use default layers with varying filter sizes
            fc_layers = [{"output_size": 512}, {"output_size": 256}]
            num_fc_layers = 2
        elif fc_layers is not None and num_fc_layers is not None:
            raise ValueError("Invalid layer parametrization, use either fc_layers or num_fc_layers only. Not both.")

        self.should_embed = should_embed
        self.embed_sequence = None

        if self.should_embed:
            logger.debug("  EmbedSequence")
            self.embed_sequence = EmbedSequence(
                vocab,
                embedding_size,
                max_sequence_length=max_sequence_length,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                embeddings_on_cpu=embeddings_on_cpu,
                dropout=dropout,
                embedding_initializer=weights_initializer,
            )

        logger.debug("  ParallelConv1D")
        in_channels = self.embed_sequence.output_shape[-1] if self.should_embed else embedding_size
        self.parallel_conv1d = ParallelConv1D(
            in_channels=in_channels,
            max_sequence_length=self.max_sequence_length,
            layers=self.conv_layers,
            default_num_filters=num_filters,
            default_filter_size=filter_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
            default_pool_function=pool_function,
            default_pool_size=pool_size,
            default_pool_padding="same",
        )

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=max_sequence_length,
            encoding_size=self.parallel_conv1d.output_shape[-1],
        )
        if self.reduce_output is not None:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.reduce_sequence.output_shape[-1],
                layers=fc_layers,
                num_layers=num_fc_layers,
                default_output_size=output_size,
                default_use_bias=use_bias,
                default_weights_initializer=weights_initializer,
                default_bias_initializer=bias_initializer,
                default_norm=norm,
                default_norm_params=norm_params,
                default_activation=activation,
                default_dropout=dropout,
            )

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented)
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_sequence = self.embed_sequence(inputs, mask=mask)
        else:
            embedded_sequence = inputs
            while len(embedded_sequence.shape) < 3:
                embedded_sequence = embedded_sequence.unsqueeze(-1)
            embedded_sequence = embedded_sequence.to(dtype=torch.float)

        # shape=(?, sequence_length, embedding_size)
        hidden = embedded_sequence

        # ================ Conv Layers ================
        hidden = self.parallel_conv1d(hidden, mask=mask)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden, mask=mask)

        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return ParallelCNNConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is not None:
            return self.fc_stack.output_shape
        return self.parallel_conv1d.output_shape


@DeveloperAPI
@register_sequence_encoder("stacked_cnn")
@register_encoder("stacked_cnn", [AUDIO, SEQUENCE, TEXT, TIMESERIES])
class StackedCNN(SequenceEncoder):
    def __init__(
        self,
        should_embed=True,
        vocab=None,
        representation="dense",
        embedding_size=256,
        max_sequence_length=None,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        conv_layers=None,
        num_conv_layers=None,
        num_filters=256,
        filter_size=5,
        strides=1,
        # todo: assess how to specify padding for equivalent to 'same'
        padding="same",
        dilation_rate=1,
        pool_function="max",
        pool_size=None,
        pool_strides=None,
        # todo: determine how to pool_padding equivalent of 'same'
        pool_padding="same",
        fc_layers=None,
        num_fc_layers=None,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        activation="relu",
        dropout=0,
        reduce_output="max",
        encoder_config=None,
        **kwargs,
    ):
        # todo: fixup docstring
        """
        :param should_embed: If True the input sequence is expected
               to be made of integers and will be mapped into embeddings
        :type should_embed: Boolean
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: Str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memroy and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param conv_layers: it is a list of dictionaries containing
               the parameters of all the convolutional layers. The length
               of the list determines the number of parallel convolutional
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `filter_size`, `num_filters`, `pool`,
               `norm` and `activation`. If any of those values
               is missing from the dictionary, the default one specified
               as a parameter of the encoder will be used instead. If both
               `conv_layers` and `num_conv_layers` are `None`, a default
               list will be assigned to `conv_layers` with the value
               `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
               {filter_size: 5}]`.
        :type conv_layers: List
        :param num_conv_layers: if `conv_layers` is `None`, this is
               the number of stacked convolutional layers.
        :type num_conv_layers: Integer
        :param filter_size:  if a `filter_size` is not already specified in
               `conv_layers` this is the default `filter_size` that
               will be used for each layer. It indicates how wide is
               the 1d convolutional filter.
        :type filter_size: Integer
        :param num_filters: if a `num_filters` is not already specified in
               `conv_layers` this is the default `num_filters` that
               will be used for each layer. It indicates the number
               of filters, and by consequence the output channels of
               the 1d convolution.
        :type num_filters: Integer
        :param pool_size: if a `pool_size` is not already specified
              in `conv_layers` this is the default `pool_size` that
              will be used for each layer. It indicates the size of
              the max pooling that will be performed along the `s` sequence
              dimension after the convolution operation.
        :type pool_size: Integer
        :param fc_layers: it is a list of dictionaries containing
               the parameters of all the fully connected layers. The length
               of the list determines the number of stacked fully connected
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `output_size`, `norm` and `activation`.
               If any of those values is missing from
               the dictionary, the default one specified as a parameter of
               the encoder will be used instead. If both `fc_layers` and
               `num_fc_layers` are `None`, a default list will be assigned
               to `fc_layers` with the value
               `[{output_size: 512}, {output_size: 256}]`
               (only applies if `reduce_output` is not `None`).
        :type fc_layers: List
        :param num_fc_layers: if `fc_layers` is `None`, this is the number
               of stacked fully connected layers (only applies if
               `reduce_output` is not `None`).
        :type num_fc_layers: Integer
        :param output_size: if a `output_size` is not already specified in
               `fc_layers` this is the default `output_size` that will be used
               for each layer. It indicates the size of the output
               of a fully connected layer.
        :type output_size: Integer
        :param norm: if a `norm` is not already specified in `conv_layers`
               or `fc_layers` this is the default `norm` that will be used
               for each layer. It indicates the norm of the output.
        :type norm: str
        :param activation: Default activation function to use
        :type activation: Str
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        :param initializer: the initializer to use. If `None` it uses
               `xavier_uniform`. Options are: `constant`, `identity`,
               `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
               `truncated_normal`, `variance_scaling`, `xavier_normal`,
               `xavier_uniform`, `xavier_normal`,
               `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
               Alternatively it is possible to specify a dictionary with
               a key `type` that identifies the type of initializer and
               other keys for its parameters,
               e.g. `{type: normal, mean: 0, stddev: 0}`.
               To know the parameters of each initializer, please refer
               to PyTorch's documentation.
        :type initializer: str
        :param reduce_output: defines how to reduce the output tensor of
               the convolutional layers along the `s` sequence length
               dimension if the rank of the tensor is greater than 2.
               Available values are: `sum`, `mean` or `avg`, `max`, `concat`
               (concatenates along the first dimension), `last` (returns
               the last vector of the first dimension) and `None` or `null`
               (which does not reduce and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        if conv_layers is not None and num_conv_layers is None:
            # use custom-defined layers
            self.conv_layers = conv_layers
            self.num_conv_layers = len(conv_layers)
        elif conv_layers is None and num_conv_layers is not None:
            # generate num_conv_layers with default parameters
            self.conv_layers = None
            self.num_conv_layers = num_conv_layers
        elif conv_layers is None and num_conv_layers is None:
            # use default layers with varying filter sizes
            self.conv_layers = [
                {
                    "filter_size": 7,
                    "pool_size": 3,
                },
                {
                    "filter_size": 7,
                    "pool_size": 3,
                },
                {
                    "filter_size": 3,
                    "pool_size": None,
                },
                {
                    "filter_size": 3,
                    "pool_size": None,
                },
                {
                    "filter_size": 3,
                    "pool_size": None,
                },
                {
                    "filter_size": 3,
                    "pool_size": 3,
                },
            ]
            self.num_conv_layers = 6
        else:
            raise ValueError("Invalid layer parametrization, use either conv_layers or " "num_conv_layers")

        # The user is expected to provide fc_layers or num_fc_layers
        # The following logic handles the case where the user either provides
        # both or neither.
        if fc_layers is None and num_fc_layers is None:
            # use default layers with varying filter sizes
            fc_layers = [{"output_size": 512}, {"output_size": 256}]
            num_fc_layers = 2
        elif fc_layers is not None and num_fc_layers is not None:
            raise ValueError("Invalid layer parametrization, use either fc_layers or " "num_fc_layers only. Not both.")

        self.max_sequence_length = max_sequence_length
        self.num_filters = num_filters
        self.should_embed = should_embed
        self.embed_sequence = None

        if self.should_embed:
            logger.debug("  EmbedSequence")
            self.embed_sequence = EmbedSequence(
                vocab,
                embedding_size,
                max_sequence_length=self.max_sequence_length,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                embeddings_on_cpu=embeddings_on_cpu,
                dropout=dropout,
                embedding_initializer=weights_initializer,
            )

        logger.debug("  Conv1DStack")
        in_channels = self.embed_sequence.output_shape[-1] if self.should_embed else embedding_size
        self.conv1d_stack = Conv1DStack(
            in_channels=in_channels,
            max_sequence_length=max_sequence_length,
            layers=self.conv_layers,
            default_num_filters=num_filters,
            default_filter_size=filter_size,
            default_strides=strides,
            default_padding=padding,
            default_dilation_rate=dilation_rate,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
            default_pool_function=pool_function,
            default_pool_size=pool_size,
            default_pool_strides=pool_strides,
            default_pool_padding=pool_padding,
        )

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=self.conv1d_stack.output_shape[-2],
            encoding_size=self.conv1d_stack.output_shape[-1],
        )
        if self.reduce_output is not None:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.reduce_sequence.output_shape[-1],
                layers=fc_layers,
                num_layers=num_fc_layers,
                default_output_size=output_size,
                default_use_bias=use_bias,
                default_weights_initializer=weights_initializer,
                default_bias_initializer=bias_initializer,
                default_norm=norm,
                default_norm_params=norm_params,
                default_activation=activation,
                default_dropout=dropout,
            )

    @staticmethod
    def get_schema_cls():
        return StackedCNNConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is None:
            return self.conv1d_stack.output_shape
        return self.fc_stack.output_shape

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented)
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_sequence = self.embed_sequence(inputs, mask=mask)
        else:
            embedded_sequence = inputs
            while len(embedded_sequence.shape) < 3:
                embedded_sequence = embedded_sequence.unsqueeze(-1)

        # shape=(?, sequence_length, embedding_size)
        hidden = embedded_sequence

        # ================ Conv Layers ================
        hidden = self.conv1d_stack(hidden, mask=mask)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden, mask=mask)

        # no reduction: hidden [batch_size, seq_size, num_filters]
        # with reduction: hidden [batch_size, output_size]
        return {"encoder_output": hidden}


@DeveloperAPI
@register_sequence_encoder("stacked_parallel_cnn")
@register_encoder("stacked_parallel_cnn", [AUDIO, SEQUENCE, TEXT, TIMESERIES])
class StackedParallelCNN(SequenceEncoder):
    def __init__(
        self,
        should_embed=True,
        vocab=None,
        representation="dense",
        embedding_size=256,
        max_sequence_length=None,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        stacked_layers=None,
        num_stacked_layers=None,
        filter_size=3,
        num_filters=256,
        pool_function="max",
        pool_size=None,
        fc_layers=None,
        num_fc_layers=None,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        activation="relu",
        dropout=0,
        reduce_output="max",
        encoder_config=None,
        **kwargs,
    ):
        # todo: review docstring
        """
        :param should_embed: If True the input sequence is expected
               to be made of integers and will be mapped into embeddings
        :type should_embed: Boolean
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: Str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memroy and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param stacked_layers: it is a of lists of list of dictionaries
               containing the parameters of the stack of
               parallel convolutional layers. The length of the list
               determines the number of stacked parallel
               convolutional layers, length of the sub-lists determines
               the number of parallel conv layers and the content
               of each dictionary determines the parameters for
               a specific layer. The available parameters for each layer are:
               `filter_size`, `num_filters`, `pool_size`, `norm` and
               `activation`. If any of those values
               is missing from the dictionary, the default one specified
               as a parameter of the encoder will be used instead. If both
               `stacked_layers` and `num_stacked_layers` are `None`,
               a default list will be assigned to `stacked_layers` with
               the value `[[{filter_size: 2}, {filter_size: 3},
               {filter_size: 4}, {filter_size: 5}], [{filter_size: 2},
               {filter_size: 3}, {filter_size: 4}, {filter_size: 5}],
               [{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
               {filter_size: 5}]]`.
        :type stacked_layers: List
        :param num_stacked_layers: if `stacked_layers` is `None`, this is
               the number of elements in the stack of
               parallel convolutional layers.
        :type num_stacked_layers: Integer
        :param filter_size:  if a `filter_size` is not already specified in
               `conv_layers` this is the default `filter_size` that
               will be used for each layer. It indicates how wide is
               the 1d convolutional filter.
        :type filter_size: Integer
        :param num_filters: if a `num_filters` is not already specified in
               `conv_layers` this is the default `num_filters` that
               will be used for each layer. It indicates the number
               of filters, and by consequence the output channels of
               the 1d convolution.
        :type num_filters: Integer
        :param pool_size: if a `pool_size` is not already specified
              in `conv_layers` this is the default `pool_size` that
              will be used for each layer. It indicates the size of
              the max pooling that will be performed along the `s` sequence
              dimension after the convolution operation.
        :type pool_size: Integer
        :param fc_layers: it is a list of dictionaries containing
               the parameters of all the fully connected layers. The length
               of the list determines the number of stacked fully connected
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `output_size`, `norm` and `activation`.
               If any of those values is missing from
               the dictionary, the default one specified as a parameter of
               the encoder will be used instead. If both `fc_layers` and
               `num_fc_layers` are `None`, a default list will be assigned
               to `fc_layers` with the value
               `[{output_size: 512}, {output_size: 256}]`
               (only applies if `reduce_output` is not `None`).
        :type fc_layers: List
        :param num_fc_layers: if `fc_layers` is `None`, this is the number
               of stacked fully connected layers (only applies if
               `reduce_output` is not `None`).
        :type num_fc_layers: Integer
        :param output_size: if a `output_size` is not already specified in
               `fc_layers` this is the default `output_size` that will be used
               for each layer. It indicates the size of the output
               of a fully connected layer.
        :type output_size: Integer
        :param norm: if a `norm` is not already specified in `conv_layers`
               or `fc_layers` this is the default `norm` that will be used
               for each layer. It indicates the norm of the output.
        :type norm: str
        :param activation: Default activation function to use
        :type activation: Str
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        :param initializer: the initializer to use. If `None` it uses
               `xavier_uniform`. Options are: `constant`, `identity`,
               `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
               `truncated_normal`, `variance_scaling`, `xavier_normal`,
               `xavier_uniform`, `xavier_normal`,
               `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
               Alternatively it is possible to specify a dictionary with
               a key `type` that identifies the type of initializer and
               other keys for its parameters,
               e.g. `{type: normal, mean: 0, stddev: 0}`.
               To know the parameters of each initializer, please refer
               to PyTorch's documentation.
        :type initializer: str
        :param reduce_output: defines how to reduce the output tensor of
               the convolutional layers along the `s` sequence length
               dimension if the rank of the tensor is greater than 2.
               Available values are: `sum`, `mean` or `avg`, `max`, `concat`
               (concatenates along the first dimension), `last` (returns
               the last vector of the first dimension) and `None` or `null`
               (which does not reduce and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.max_sequence_length = max_sequence_length
        self.embedding_size = embedding_size

        if stacked_layers is not None and num_stacked_layers is None:
            # use custom-defined layers
            self.stacked_layers = stacked_layers
            self.num_stacked_layers = len(stacked_layers)
        elif stacked_layers is None and num_stacked_layers is not None:
            # generate num_conv_layers with default parameters
            self.stacked_layers = None
            self.num_stacked_layers = num_stacked_layers
        elif stacked_layers is None and num_stacked_layers is None:
            # use default layers with varying filter sizes
            self.stacked_layers = [
                [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}],
                [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}],
                [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}],
            ]
            self.num_stacked_layers = 6
        else:
            raise ValueError("Invalid layer parametrization, use either stacked_layers or" " num_stacked_layers")

        # The user is expected to provide fc_layers or num_fc_layers
        # The following logic handles the case where the user either provides
        # both or neither.
        if fc_layers is None and num_fc_layers is None:
            # use default layers with varying filter sizes
            fc_layers = [{"output_size": 512}, {"output_size": 256}]
            num_fc_layers = 2
        elif fc_layers is not None and num_fc_layers is not None:
            raise ValueError("Invalid layer parametrization, use either fc_layers or " "num_fc_layers only. Not both.")

        self.should_embed = should_embed
        self.embed_sequence = None

        if self.should_embed:
            logger.debug("  EmbedSequence")
            self.embed_sequence = EmbedSequence(
                vocab,
                embedding_size,
                max_sequence_length=self.max_sequence_length,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                embeddings_on_cpu=embeddings_on_cpu,
                dropout=dropout,
                embedding_initializer=weights_initializer,
            )

        in_channels = self.embed_sequence.output_shape[-1] if self.should_embed else embedding_size
        logger.debug("  ParallelConv1DStack")
        self.parallel_conv1d_stack = ParallelConv1DStack(
            in_channels=in_channels,
            stacked_layers=self.stacked_layers,
            max_sequence_length=max_sequence_length,
            default_num_filters=num_filters,
            default_filter_size=filter_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
            default_pool_function=pool_function,
            default_pool_size=pool_size,
        )

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=self.parallel_conv1d_stack.output_shape[-2],
            encoding_size=self.parallel_conv1d_stack.output_shape[-1],
        )
        if self.reduce_output is not None:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.reduce_sequence.output_shape[-1],
                layers=fc_layers,
                num_layers=num_fc_layers,
                default_output_size=output_size,
                default_use_bias=use_bias,
                default_weights_initializer=weights_initializer,
                default_bias_initializer=bias_initializer,
                default_norm=norm,
                default_norm_params=norm_params,
                default_activation=activation,
                default_dropout=dropout,
            )

    @staticmethod
    def get_schema_cls():
        return StackedParallelCNNConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is not None:
            return self.fc_stack.output_shape
        return self.parallel_conv1d_stack.output_shape

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented)
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_sequence = self.embed_sequence(inputs, mask=mask)
        else:
            embedded_sequence = inputs
            while len(embedded_sequence.shape) < 3:
                embedded_sequence = embedded_sequence.unsqueeze(-1)

        # shape=(?, sequence_length, embedding_size)
        hidden = embedded_sequence

        # ================ Conv Layers ================
        hidden = self.parallel_conv1d_stack(hidden, mask=mask)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden, mask=mask)

        # no reduction: hidden [batch_size, seq_size, num_filter]
        # with reduction: hidden [batch_size, output_size]
        return {"encoder_output": hidden}


@DeveloperAPI
@register_sequence_encoder("rnn")
@register_encoder("rnn", [AUDIO, SEQUENCE, TEXT, TIMESERIES])
class StackedRNN(SequenceEncoder):
    def __init__(
        self,
        should_embed=True,
        vocab=None,
        representation="dense",
        embedding_size=256,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        num_layers=1,
        max_sequence_length=None,
        state_size=256,
        cell_type="rnn",
        bidirectional=False,
        activation="tanh",
        recurrent_activation="sigmoid",
        unit_forget_bias=True,
        recurrent_initializer="orthogonal",
        dropout=0.0,
        recurrent_dropout=0.0,
        fc_layers=None,
        num_fc_layers=0,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        fc_activation="relu",
        fc_dropout=0,
        reduce_output="last",
        encoder_config=None,
        **kwargs,
    ):
        # todo: fix up docstring
        """
        :param should_embed: If True the input sequence is expected
               to be made of integers and will be mapped into embeddings
        :type should_embed: Boolean
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: Str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memroy and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param conv_layers: it is a list of dictionaries containing
               the parameters of all the convolutional layers. The length
               of the list determines the number of parallel convolutional
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `filter_size`, `num_filters`, `pool`,
               `norm`, `activation` and `regularize`. If any of those values
               is missing from the dictionary, the default one specified
               as a parameter of the encoder will be used instead. If both
               `conv_layers` and `num_conv_layers` are `None`, a default
               list will be assigned to `conv_layers` with the value
               `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
               {filter_size: 5}]`.
        :type conv_layers: List
        :param num_conv_layers: if `conv_layers` is `None`, this is
               the number of stacked convolutional layers.
        :type num_conv_layers: Integer
        :param filter_size:  if a `filter_size` is not already specified in
               `conv_layers` this is the default `filter_size` that
               will be used for each layer. It indicates how wide is
               the 1d convolutional filter.
        :type filter_size: Integer
        :param num_filters: if a `num_filters` is not already specified in
               `conv_layers` this is the default `num_filters` that
               will be used for each layer. It indicates the number
               of filters, and by consequence the output channels of
               the 1d convolution.
        :type num_filters: Integer
        :param pool_size: if a `pool_size` is not already specified
              in `conv_layers` this is the default `pool_size` that
              will be used for each layer. It indicates the size of
              the max pooling that will be performed along the `s` sequence
              dimension after the convolution operation.
        :type pool_size: Integer
        :param num_rec_layers: the number of stacked recurrent layers.
        :type num_rec_layers: Integer
        :param cell_type: the type of recurrent cell to use.
               Available values are: `rnn`, `lstm`, `gru`.
               For reference about the differences between the cells please
               refer to PyTorch's documentation.
        :type cell_type: str
        :param state_size: the size of the state of the rnn.
        :type state_size: Integer
        :param bidirectional: if `True` two recurrent networks will perform
               encoding in the forward and backward direction and
               their outputs will be concatenated.
        :type bidirectional: Boolean
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        :param recurrent_dropout: Dropout rate for the recurrent stack.
        :type recurrent_dropout: float
        :param initializer: the initializer to use. If `None` it uses
               `xavier_uniform`. Options are: `constant`, `identity`,
               `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
               `truncated_normal`, `variance_scaling`, `xavier_normal`,
               `xavier_uniform`, `xavier_normal`,
               `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
               Alternatively it is possible to specify a dictionary with
               a key `type` that identifies the type of initializer and
               other keys for its parameters,
               e.g. `{type: normal, mean: 0, stddev: 0}`.
               To know the parameters of each initializer, please refer
               to PyTorch's documentation.
        :type initializer: str
        :param reduce_output: defines how to reduce the output tensor of
               the convolutional layers along the `s` sequence length
               dimension if the rank of the tensor is greater than 2.
               Available values are: `sum`, `mean` or `avg`, `max`, `concat`
               (concatenates along the first dimension), `last` (returns
               the last vector of the first dimension) and `None` or `null`
               (which does not reduce and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.max_sequence_length = max_sequence_length
        self.hidden_size = state_size
        self.embedding_size = embedding_size

        self.should_embed = should_embed
        self.embed_sequence = None

        if self.should_embed:
            logger.debug("  EmbedSequence")
            self.embed_sequence = EmbedSequence(
                vocab,
                embedding_size,
                max_sequence_length=self.max_sequence_length,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                embeddings_on_cpu=embeddings_on_cpu,
                dropout=dropout,
                embedding_initializer=weights_initializer,
            )

        logger.debug("  RecurrentStack")
        input_size = self.embed_sequence.output_shape[-1] if self.should_embed else embedding_size
        self.recurrent_stack = RecurrentStack(
            input_size=input_size,
            hidden_size=state_size,
            cell_type=cell_type,
            max_sequence_length=max_sequence_length,
            num_layers=num_layers,
            bidirectional=bidirectional,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            unit_forget_bias=unit_forget_bias,
            weights_initializer=weights_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            dropout=recurrent_dropout,
        )

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=self.recurrent_stack.output_shape[-2],
            encoding_size=self.recurrent_stack.output_shape[-1],  # state_size
        )
        if self.reduce_output is None:
            self.supports_masking = True
        else:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.reduce_sequence.output_shape[-1],
                layers=fc_layers,
                num_layers=num_fc_layers,
                default_output_size=output_size,
                default_use_bias=use_bias,
                default_weights_initializer=weights_initializer,
                default_bias_initializer=bias_initializer,
                default_norm=norm,
                default_norm_params=norm_params,
                default_activation=fc_activation,
                default_dropout=fc_dropout,
            )

    @staticmethod
    def get_schema_cls():
        return StackedRNNConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is not None:
            return self.fc_stack.output_shape
        return self.recurrent_stack.output_shape

    def input_dtype(self):
        return torch.int32

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented)
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_sequence = self.embed_sequence(inputs, mask=mask)
        else:
            embedded_sequence = inputs
            while len(embedded_sequence.shape) < 3:
                embedded_sequence = embedded_sequence.unsqueeze(-1)

        # shape=(?, sequence_length, embedding_size)
        hidden = embedded_sequence

        # ================ Recurrent Layers ================
        hidden, final_state = self.recurrent_stack(hidden, mask=mask)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden, mask=mask)

        return {"encoder_output": hidden, "encoder_output_state": final_state}


@DeveloperAPI
@register_sequence_encoder("cnnrnn")
@register_encoder("cnnrnn", [AUDIO, SEQUENCE, TEXT, TIMESERIES])
class StackedCNNRNN(SequenceEncoder):
    def __init__(
        self,
        should_embed=True,
        vocab=None,
        max_sequence_length=None,
        representation="dense",
        embedding_size=256,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        conv_layers=None,
        num_conv_layers=None,
        num_filters=256,
        filter_size=5,
        strides=1,
        padding="same",
        dilation_rate=1,
        conv_activation="relu",
        conv_dropout=0.0,
        pool_function="max",
        pool_size=2,
        pool_strides=None,
        pool_padding="same",
        num_rec_layers=1,
        state_size=256,
        cell_type="rnn",
        bidirectional=False,
        activation="tanh",
        recurrent_activation="sigmoid",
        unit_forget_bias=True,
        recurrent_initializer="orthogonal",
        dropout=0.0,
        recurrent_dropout=0.0,
        fc_layers=None,
        num_fc_layers=0,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        fc_activation="relu",
        fc_dropout=0,
        reduce_output="last",
        encoder_config=None,
        **kwargs,
    ):
        # todo: fix up docstring
        """
        :param should_embed: If True the input sequence is expected
               to be made of integers and will be mapped into embeddings
        :type should_embed: Boolean
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: Str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memroy and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param num_layers: the number of stacked recurrent layers.
        :type num_layers: Integer
        :param cell_type: the type of recurrent cell to use.
               Available values are: `rnn`, `lstm`, `gru`.
               For reference about the differences between the cells please
               refer to PyTorch's documentation.
        :type cell_type: str
        :param state_size: the size of the state of the rnn.
        :type state_size: Integer
        :param bidirectional: if `True` two recurrent networks will perform
               encoding in the forward and backward direction and
               their outputs will be concatenated.
        :type bidirectional: Boolean
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        :param recurrent_dropout: Dropout rate for the recurrent stack.
        :type recurrent_dropout: float
        :param initializer: the initializer to use. If `None` it uses
               `xavier_uniform`. Options are: `constant`, `identity`,
               `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
               `truncated_normal`, `variance_scaling`, `xavier_normal`,
               `xavier_uniform`, `xavier_normal`,
               `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
               Alternatively it is possible to specify a dictionary with
               a key `type` that identifies the type of initializer and
               other keys for its parameters,
               e.g. `{type: normal, mean: 0, stddev: 0}`.
               To know the parameters of each initializer, please refer
               to PyTorch's documentation.
        :type initializer: str
        :param reduce_output: defines how to reduce the output tensor of
               the convolutional layers along the `s` sequence length
               dimension if the rank of the tensor is greater than 2.
               Available values are: `sum`, `mean` or `avg`, `max`, `concat`
               (concatenates along the first dimension), `last` (returns
               the last vector of the first dimension) and `None` or `null`
               (which does not reduce and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        if conv_layers is not None and num_conv_layers is None:
            # use custom-defined layers
            self.conv_layers = conv_layers
            self.num_conv_layers = len(conv_layers)
        elif conv_layers is None and num_conv_layers is not None:
            # generate num_conv_layers with default parameters
            self.conv_layers = None
            self.num_conv_layers = num_conv_layers
        elif conv_layers is None and num_conv_layers is None:
            # use default layers with varying filter sizes
            self.conv_layers = [{"pool_size": 3}, {"pool_size": None}]
            self.num_conv_layers = 2
        else:
            raise ValueError("Invalid layer parametrization, use either conv_layers or " "num_conv_layers")

        self.max_sequence_length = max_sequence_length
        self.should_embed = should_embed
        self.embed_sequence = None

        if self.should_embed:
            logger.debug("  EmbedSequence")
            self.embed_sequence = EmbedSequence(
                vocab,
                embedding_size,
                max_sequence_length=self.max_sequence_length,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                embeddings_on_cpu=embeddings_on_cpu,
                dropout=dropout,
                embedding_initializer=weights_initializer,
            )

        logger.debug("  Conv1DStack")
        in_channels = self.embed_sequence.output_shape[-1] if self.should_embed else embedding_size
        self.conv1d_stack = Conv1DStack(
            in_channels=in_channels,
            max_sequence_length=max_sequence_length,
            layers=self.conv_layers,
            default_num_filters=num_filters,
            default_filter_size=filter_size,
            default_strides=strides,
            default_padding=padding,
            default_dilation_rate=dilation_rate,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=conv_activation,
            default_dropout=conv_dropout,
            default_pool_function=pool_function,
            default_pool_size=pool_size,
            default_pool_strides=pool_strides,
            default_pool_padding=pool_padding,
        )

        logger.debug("  RecurrentStack")
        self.recurrent_stack = RecurrentStack(
            input_size=self.conv1d_stack.output_shape[1],
            hidden_size=state_size,
            max_sequence_length=self.conv1d_stack.output_shape[0],
            cell_type=cell_type,
            num_layers=num_rec_layers,
            bidirectional=bidirectional,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            unit_forget_bias=unit_forget_bias,
            weights_initializer=weights_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            dropout=recurrent_dropout,
        )

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=self.recurrent_stack.output_shape[-2],
            encoding_size=self.recurrent_stack.output_shape[-1],  # State size
        )
        if self.reduce_output is not None:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.reduce_sequence.output_shape[-1],
                layers=fc_layers,
                num_layers=num_fc_layers,
                default_output_size=output_size,
                default_use_bias=use_bias,
                default_weights_initializer=weights_initializer,
                default_bias_initializer=bias_initializer,
                default_norm=norm,
                default_norm_params=norm_params,
                default_activation=fc_activation,
                default_dropout=fc_dropout,
            )

    @staticmethod
    def get_schema_cls():
        return StackedCNNRNNConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is not None:
            return self.fc_stack.output_shape
        return self.recurrent_stack.output_shape

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented)
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_sequence = self.embed_sequence(inputs, mask=mask)
        else:
            embedded_sequence = inputs
            while len(embedded_sequence.shape) < 3:
                embedded_sequence = embedded_sequence.unsqueeze(-1)

        # shape=(?, sequence_length, embedding_size)
        hidden = embedded_sequence

        # ================ Conv Layers ================
        hidden = self.conv1d_stack(hidden, mask=mask)

        # ================ Recurrent Layers ================
        hidden, final_state = self.recurrent_stack(hidden)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden, mask=mask)

        # no reduction: hidden [batch_size, seq_size, state_size]
        # with reduction: hidden [batch_size, seq_size, output_size]
        # final_state: if rnn/gru [batch_size, state_size]
        #              lstm ([batch_size, state_size], [batch_size, state_size])
        return {"encoder_output": hidden, "encoder_output_state": final_state}


@DeveloperAPI
@register_encoder("transformer", [SEQUENCE, TEXT, TIMESERIES])
class StackedTransformer(SequenceEncoder):
    def __init__(
        self,
        max_sequence_length,
        should_embed=True,
        vocab=None,
        representation="dense",
        embedding_size=256,
        embeddings_trainable=True,
        pretrained_embeddings=None,
        embeddings_on_cpu=False,
        num_layers=1,
        hidden_size=256,
        num_heads=8,
        transformer_output_size=256,
        dropout=0.1,
        fc_layers=None,
        num_fc_layers=0,
        output_size=256,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        fc_activation="relu",
        fc_dropout=0,
        reduce_output="last",
        encoder_config=None,
        **kwargs,
    ):
        # todo: update docstring as needed
        """
        :param should_embed: If True the input sequence is expected
               to be made of integers and will be mapped into embeddings
        :type should_embed: Boolean
        :param vocab: Vocabulary of the input feature to encode
        :type vocab: List
        :param representation: the possible values are `dense` and `sparse`.
               `dense` means the embeddings are initialized randomly,
               `sparse` means they are initialized to be one-hot encodings.
        :type representation: Str (one of 'dense' or 'sparse')
        :param embedding_size: it is the maximum embedding size, the actual
               size will be `min(vocabulary_size, embedding_size)`
               for `dense` representations and exactly `vocabulary_size`
               for the `sparse` encoding, where `vocabulary_size` is
               the number of different strings appearing in the training set
               in the column the feature is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_trainable: If `True` embeddings are trained during
               the training process, if `False` embeddings are fixed.
               It may be useful when loading pretrained embeddings
               for avoiding finetuning them. This parameter has effect only
               for `representation` is `dense` as `sparse` one-hot encodings
                are not trainable.
        :type embeddings_trainable: Boolean
        :param pretrained_embeddings: by default `dense` embeddings
               are initialized randomly, but this parameter allows to specify
               a path to a file containing embeddings in the GloVe format.
               When the file containing the embeddings is loaded, only the
               embeddings with labels present in the vocabulary are kept,
               the others are discarded. If the vocabulary contains strings
               that have no match in the embeddings file, their embeddings
               are initialized with the average of all other embedding plus
               some random noise to make them different from each other.
               This parameter has effect only if `representation` is `dense`.
        :type pretrained_embeddings: str (filepath)
        :param embeddings_on_cpu: by default embeddings matrices are stored
               on GPU memory if a GPU is used, as it allows
               for faster access, but in some cases the embedding matrix
               may be really big and this parameter forces the placement
               of the embedding matrix in regular memroy and the CPU is used
               to resolve them, slightly slowing down the process
               as a result of data transfer between CPU and GPU memory.
        :param conv_layers: it is a list of dictionaries containing
               the parameters of all the convolutional layers. The length
               of the list determines the number of parallel convolutional
               layers and the content of each dictionary determines
               the parameters for a specific layer. The available parameters
               for each layer are: `filter_size`, `num_filters`, `pool`,
               `norm`, `activation` and `regularize`. If any of those values
               is missing from the dictionary, the default one specified
               as a parameter of the encoder will be used instead. If both
               `conv_layers` and `num_conv_layers` are `None`, a default
               list will be assigned to `conv_layers` with the value
               `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
               {filter_size: 5}]`.
        :type conv_layers: List
        :param num_conv_layers: if `conv_layers` is `None`, this is
               the number of stacked convolutional layers.
        :type num_conv_layers: Integer
        :param filter_size:  if a `filter_size` is not already specified in
               `conv_layers` this is the default `filter_size` that
               will be used for each layer. It indicates how wide is
               the 1d convolutional filter.
        :type filter_size: Integer
        :param num_filters: if a `num_filters` is not already specified in
               `conv_layers` this is the default `num_filters` that
               will be used for each layer. It indicates the number
               of filters, and by consequence the output channels of
               the 1d convolution.
        :type num_filters: Integer
        :param pool_size: if a `pool_size` is not already specified
              in `conv_layers` this is the default `pool_size` that
              will be used for each layer. It indicates the size of
              the max pooling that will be performed along the `s` sequence
              dimension after the convolution operation.
        :type pool_size: Integer
        :param num_rec_layers: the number of stacked recurrent layers.
        :type num_rec_layers: Integer
        :param cell_type: the type of recurrent cell to use.
               Available values are: `rnn`, `lstm`, `lstm_block`, `lstm`,
               `ln`, `lstm_cudnn`, `gru`, `gru_block`, `gru_cudnn`.
               For reference about the differences between the cells please
               refer to PyTorch's documentation. We suggest to use the
               `block` variants on CPU and the `cudnn` variants on GPU
               because of their increased speed.
        :type cell_type: str
        :param state_size: the size of the state of the rnn.
        :type state_size: Integer
        :param bidirectional: if `True` two recurrent networks will perform
               encoding in the forward and backward direction and
               their outputs will be concatenated.
        :type bidirectional: Boolean
        :param dropout: determines if there should be a dropout layer before
               returning the encoder output.
        :type dropout: Boolean
        :param initializer: the initializer to use. If `None` it uses
               `xavier_uniform`. Options are: `constant`, `identity`,
               `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
               `truncated_normal`, `variance_scaling`, `xavier_normal`,
               `xavier_uniform`, `xavier_normal`,
               `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
               Alternatively it is possible to specify a dictionary with
               a key `type` that identifies the type of initializer and
               other keys for its parameters,
               e.g. `{type: normal, mean: 0, stddev: 0}`.
               To know the parameters of each initializer, please refer
               to PyTorch's documentation.
        :type initializer: str
        :param reduce_output: defines how to reduce the output tensor of
               the convolutional layers along the `s` sequence length
               dimension if the rank of the tensor is greater than 2.
               Available values are: `sum`, `mean` or `avg`, `max`, `concat`
               (concatenates along the first dimension), `last` (returns
               the last vector of the first dimension) and `None` or `null`
               (which does not reduce and returns the full tensor).
        :type reduce_output: str
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.max_sequence_length = max_sequence_length

        self.should_embed = should_embed
        self.should_project = False
        self.embed_sequence = None

        if self.should_embed:
            logger.debug("  EmbedSequence")
            self.embed_sequence = TokenAndPositionEmbedding(
                max_sequence_length=max_sequence_length,
                vocab=vocab,
                embedding_size=embedding_size,
                representation=representation,
                embeddings_trainable=embeddings_trainable,
                pretrained_embeddings=pretrained_embeddings,
                embeddings_on_cpu=embeddings_on_cpu,
                dropout=dropout,
                embedding_initializer=weights_initializer,
            )
            # If vocab size is smaller than embedding size, embedding layer will use len(vocab) as embedding_size.
            used_embedding_size = self.embed_sequence.output_shape[-1]
            if used_embedding_size != hidden_size:
                logger.debug("  project_to_embed_size")
                self.project_to_hidden_size = nn.Linear(self.embed_sequence.output_shape[-1], hidden_size)
                self.should_project = True
        else:
            logger.debug("  project_to_embed_size")
            self.project_to_hidden_size = nn.Linear(1, hidden_size)
            self.should_project = True

        logger.debug("  TransformerStack")
        self.transformer_stack = TransformerStack(
            input_size=hidden_size,
            max_sequence_length=max_sequence_length,
            hidden_size=hidden_size,
            num_heads=num_heads,
            output_size=transformer_output_size,
            num_layers=num_layers,
            dropout=dropout,
        )

        self.reduce_output = reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=reduce_output,
            max_sequence_length=self.transformer_stack.output_shape[-2],
            encoding_size=self.transformer_stack.output_shape[-1],  # hidden_size
        )
        if self.reduce_output is None:
            self.supports_masking = True
        else:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.reduce_sequence.output_shape[-1],
                layers=fc_layers,
                num_layers=num_fc_layers,
                default_output_size=output_size,
                default_use_bias=use_bias,
                default_weights_initializer=weights_initializer,
                default_bias_initializer=bias_initializer,
                default_norm=norm,
                default_norm_params=norm_params,
                default_activation=fc_activation,
                default_dropout=fc_dropout,
            )

    @staticmethod
    def get_schema_cls():
        return StackedTransformerConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.max_sequence_length])

    @property
    def output_shape(self) -> torch.Size:
        if self.reduce_output is not None:
            return self.fc_stack.output_shape
        return self.transformer_stack.output_shape

    def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None):
        """
        :param inputs: The input sequence fed into the encoder.
               Shape: [batch x sequence length], type torch.int32
        :param mask: Input mask (unused, not yet implemented)
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_sequence = self.embed_sequence(inputs, mask=mask)
        else:
            embedded_sequence = inputs
            while len(embedded_sequence.shape) < 3:
                embedded_sequence = embedded_sequence.unsqueeze(-1)

        # shape=(?, sequence_length, embedding_size)
        if self.should_project:
            hidden = self.project_to_hidden_size(embedded_sequence)
        else:
            hidden = embedded_sequence
        # shape=(?, sequence_length, hidden)

        # ================ Transformer Layers ================
        hidden = self.transformer_stack(hidden, mask=mask)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden, mask=mask)

        return {"encoder_output": hidden}



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import IMAGE
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.convolutional_modules import Conv2DStack, ResNet
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.modules.mlp_mixer_modules import MLPMixer
from ludwig.schema.encoders.image.base import MLPMixerConfig, ResNetConfig, Stacked2DCNNConfig, ViTConfig
from ludwig.utils.torch_utils import FreezeModule

logger = logging.getLogger(__name__)


@DeveloperAPI
class ImageEncoder(Encoder):
    pass


@DeveloperAPI
@register_encoder("stacked_cnn", IMAGE)
class Stacked2DCNN(ImageEncoder):
    def __init__(
        self,
        height: int,
        width: int,
        conv_layers: Optional[List[Dict]] = None,
        num_conv_layers: Optional[int] = None,
        num_channels: int = None,
        out_channels: int = 32,
        kernel_size: Union[int, Tuple[int]] = 3,
        stride: Union[int, Tuple[int]] = 1,
        padding: Union[int, Tuple[int], str] = "valid",
        dilation: Union[int, Tuple[int]] = 1,
        conv_use_bias: bool = True,
        padding_mode: str = "zeros",
        conv_norm: Optional[str] = None,
        conv_norm_params: Optional[Dict[str, Any]] = None,
        conv_activation: str = "relu",
        conv_dropout: int = 0,
        pool_function: str = "max",
        pool_kernel_size: Union[int, Tuple[int]] = 2,
        pool_stride: Union[int, Tuple[int]] = None,
        pool_padding: Union[int, Tuple[int]] = 0,
        pool_dilation: Union[int, Tuple[int]] = 1,
        groups: int = 1,
        fc_layers: Optional[List[Dict]] = None,
        num_fc_layers: Optional[int] = 1,
        output_size: int = 128,
        fc_use_bias: bool = True,
        fc_weights_initializer: str = "xavier_uniform",
        fc_bias_initializer: str = "zeros",
        fc_norm: Optional[str] = None,
        fc_norm_params: Optional[Dict[str, Any]] = None,
        fc_activation: str = "relu",
        fc_dropout: float = 0,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        # map parameter input feature config names to internal names
        img_height = height
        img_width = width
        first_in_channels = num_channels

        self._input_shape = (first_in_channels, img_height, img_width)

        if first_in_channels is None:
            raise ValueError("first_in_channels must not be None.")

        logger.debug("  Conv2DStack")
        self.conv_stack_2d = Conv2DStack(
            img_height=img_height,
            img_width=img_width,
            layers=conv_layers,
            num_layers=num_conv_layers,
            first_in_channels=first_in_channels,
            default_out_channels=out_channels,
            default_kernel_size=kernel_size,
            default_stride=stride,
            default_padding=padding,
            default_dilation=dilation,
            default_groups=groups,
            default_use_bias=conv_use_bias,
            default_padding_mode=padding_mode,
            default_norm=conv_norm,
            default_norm_params=conv_norm_params,
            default_activation=conv_activation,
            default_dropout=conv_dropout,
            default_pool_function=pool_function,
            default_pool_kernel_size=pool_kernel_size,
            default_pool_stride=pool_stride,
            default_pool_padding=pool_padding,
            default_pool_dilation=pool_dilation,
        )
        out_channels, img_height, img_width = self.conv_stack_2d.output_shape
        first_fc_layer_input_size = out_channels * img_height * img_width

        self.flatten = torch.nn.Flatten()

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=first_fc_layer_input_size,
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=fc_use_bias,
            default_weights_initializer=fc_weights_initializer,
            default_bias_initializer=fc_bias_initializer,
            default_norm=fc_norm,
            default_norm_params=fc_norm_params,
            default_activation=fc_activation,
            default_dropout=fc_dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        :param inputs: The inputs fed into the encoder.
                Shape: [batch x channels x height x width], type torch.uint8
        """

        hidden = self.conv_stack_2d(inputs)
        hidden = self.flatten(hidden)
        outputs = self.fc_stack(hidden)

        return {"encoder_output": outputs}

    @staticmethod
    def get_schema_cls():
        return Stacked2DCNNConfig

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)


@DeveloperAPI
@register_encoder("_resnet_legacy", IMAGE)
class ResNetEncoder(ImageEncoder):
    def __init__(
        self,
        height: int,
        width: int,
        resnet_size: int = 50,
        num_channels: int = 3,
        out_channels: int = 16,
        kernel_size: Union[int, Tuple[int]] = 3,
        conv_stride: Union[int, Tuple[int]] = 1,
        first_pool_kernel_size: Union[int, Tuple[int]] = None,
        first_pool_stride: Union[int, Tuple[int]] = None,
        batch_norm_momentum: float = 0.1,
        batch_norm_epsilon: float = 0.001,
        fc_layers: Optional[List[Dict]] = None,
        num_fc_layers: Optional[int] = 1,
        output_size: int = 256,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict[str, Any]] = None,
        activation: str = "relu",
        dropout: float = 0,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")
        # map parameter input feature config names to internal names
        img_height = height
        img_width = width
        first_in_channels = num_channels

        self._input_shape = (first_in_channels, img_height, img_width)

        logger.debug("  ResNet")
        self.resnet = ResNet(
            img_height=img_height,
            img_width=img_width,
            first_in_channels=first_in_channels,
            out_channels=out_channels,
            resnet_size=resnet_size,
            kernel_size=kernel_size,
            conv_stride=conv_stride,
            first_pool_kernel_size=first_pool_kernel_size,
            first_pool_stride=first_pool_stride,
            batch_norm_momentum=batch_norm_momentum,
            batch_norm_epsilon=batch_norm_epsilon,
        )
        first_fc_layer_input_size = self.resnet.output_shape[0]

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=first_fc_layer_input_size,
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        hidden = self.resnet(inputs)
        axes = [2, 3]
        hidden = torch.mean(hidden, axes)
        hidden = self.fc_stack(hidden)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return ResNetConfig

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)


@DeveloperAPI
@register_encoder("mlp_mixer", IMAGE)
class MLPMixerEncoder(ImageEncoder):
    def __init__(
        self,
        height: int,
        width: int,
        num_channels: int = None,
        patch_size: int = 16,
        embed_size: int = 512,
        token_size: int = 2048,
        channel_dim: int = 256,
        num_layers: int = 8,
        dropout: float = 0.0,
        avg_pool: bool = True,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")
        # map parameter input feature config names to internal names
        img_height = height
        img_width = width
        in_channels = num_channels

        if num_channels is None:
            raise RuntimeError("num_channels must not be None")

        self._input_shape = (in_channels, img_height, img_width)

        logger.debug("  MLPMixer")
        self.mlp_mixer = MLPMixer(
            img_height=img_height,
            img_width=img_width,
            in_channels=in_channels,
            patch_size=patch_size,
            embed_size=embed_size,
            token_size=token_size,
            channel_dim=channel_dim,
            num_layers=num_layers,
            dropout=dropout,
            avg_pool=avg_pool,
        )

        self._output_shape = self.mlp_mixer.output_shape

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        hidden = self.mlp_mixer(inputs)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return MLPMixerConfig

    @property
    def input_shape(self) -> torch.Size:
        return self._input_shape

    @property
    def output_shape(self) -> torch.Size:
        return self._output_shape


@DeveloperAPI
@register_encoder("_vit_legacy", IMAGE)
class ViTEncoder(ImageEncoder):
    def __init__(
        self,
        height: int,
        width: int,
        num_channels: int = 3,
        use_pretrained: bool = True,
        pretrained_model: str = "google/vit-base-patch16-224",
        saved_weights_in_checkpoint: bool = False,
        hidden_size: int = 768,
        num_hidden_layers: int = 12,
        num_attention_heads: int = 12,
        intermediate_size: int = 3072,
        hidden_act: str = "gelu",
        hidden_dropout_prob: float = 0.1,
        attention_probs_dropout_prob: float = 0.1,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        gradient_checkpointing: bool = False,
        patch_size: int = 16,
        trainable: bool = True,
        output_attentions: bool = False,
        encoder_config=None,
        **kwargs,
    ):
        """Creates a ViT encoder using transformers.ViTModel.

        use_pretrained: If True, uses a pretrained transformer based on the
            pretrained_model argument.
        pretrained: If str, expects the path to a pretrained model or the id of
            a model on huggingface.co, and ignores the configuration provided in
            the arguments.
        """
        super().__init__()
        self.config = encoder_config

        try:
            from transformers import ViTConfig, ViTModel
        except ModuleNotFoundError:
            raise RuntimeError(
                " transformers is not installed. "
                "In order to install all image feature dependencies run "
                "pip install ludwig[image]"
            )

        # map parameter input feature config names to internal names
        img_height = height
        img_width = width
        in_channels = num_channels

        img_width = img_width or img_height
        if img_width != img_height:
            raise ValueError("img_height and img_width should be identical.")
        self._input_shape = (in_channels, img_height, img_width)

        if use_pretrained and not saved_weights_in_checkpoint:
            transformer = ViTModel.from_pretrained(pretrained_model)
        else:
            config = ViTConfig(
                image_size=img_height,
                num_channels=in_channels,
                patch_size=patch_size,
                hidden_size=hidden_size,
                num_hidden_layers=num_hidden_layers,
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                hidden_act=hidden_act,
                hidden_dropout_prob=hidden_dropout_prob,
                attention_probs_dropout_prob=attention_probs_dropout_prob,
                initializer_range=initializer_range,
                layer_norm_eps=layer_norm_eps,
                gradient_checkpointing=gradient_checkpointing,
            )
            transformer = ViTModel(config)

        self.transformer = FreezeModule(transformer, frozen=not trainable)

        self._output_shape = (transformer.config.hidden_size,)
        self.output_attentions = output_attentions

    def forward(self, inputs: torch.Tensor, head_mask: torch.Tensor = None) -> Dict[str, torch.Tensor]:
        output = self.transformer.module(inputs, head_mask=head_mask, output_attentions=self.output_attentions)
        return_dict = {"encoder_output": output.pooler_output}
        if self.output_attentions:
            return_dict["attentions"] = output.attentions
        return return_dict

    @staticmethod
    def get_schema_cls():
        return ViTConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Dict, List, Optional

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import DATE
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.embedding_modules import Embed
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.schema.encoders.date_encoders import DateEmbedConfig, DateWaveConfig
from ludwig.utils import torch_utils

logger = logging.getLogger(__name__)

# Year, month, day, weekday, yearday, hour, minute, seconds, second_of_day.
# TODO: Share this constant with date_feature.DATE_VECTOR_SIZE.
DATE_INPUT_SIZE = 9


@DeveloperAPI
@register_encoder("embed", DATE)
class DateEmbed(Encoder):
    def __init__(
        self,
        embedding_size: int = 10,
        embeddings_on_cpu: bool = False,
        fc_layers: Optional[List[Dict]] = None,
        num_fc_layers: int = 0,
        output_size: int = 10,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict] = None,
        activation: str = "relu",
        dropout: float = 0,
        encoder_config=None,
        **kwargs,
    ):
        """
        :param embedding_size: The maximum embedding size, the actual size
            will be `min(vocabulary_size, embedding_size)` for `dense`
            representations and exactly `vocabulary_size` for the `sparse`
            encoding, where `vocabulary_size` is the number of different
            strings appearing in the training set in the column the feature
            is named after (plus 1 for `<UNK>`).
        :type embedding_size: Integer
        :param embeddings_on_cpu: by default embeddings matrices are stored
            on GPU memory if a GPU is used, as it allows for faster access,
            but in some cases the embedding matrix may be really big and
            this parameter forces the placement of the embedding matrix in
            regular memory and the CPU is used to resolve them, slightly
            slowing down the process as a result of data transfer between
            CPU and GPU memory.
        :param fc_layers: list of dictionaries containing the parameters of
            all the fully connected layers.
        :type fc_layers: List
        :param num_fc_layers: Number of stacked fully connected layers.
        :type num_fc_layers: Integer
        :param output_size: Size of each layer.
        :type output_size: Integer
        :param use_bias: bool determines where to use a bias vector.
        :type use_bias: bool
        :param weights_initializer: Initializer for the weights (aka kernel)
            matrix.
        :type weights_initializer: string
        :param bias_initializer: Initializer for the bias vector.
        :type bias_initializer: string
        :param norm: type of normalization to use 'batch' or 'layer'.
        :type norm: string, default None
        :param norm_params: parameters to pass to normalization function.
        :type norm_params: dictionary
        :param activation: Activation function to use.
        :type activation: string
        :param dropout: determines if there should be a dropout layer before
            returning the encoder output.
        :type dropout: float
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        logger.debug("  year FCStack")
        self.year_fc = FCStack(
            first_layer_input_size=1,
            num_layers=1,
            default_output_size=1,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=None,
            default_norm_params=None,
            default_activation=None,
            default_dropout=dropout,
        )

        logger.debug("  month Embed")
        self.embed_month = Embed(
            [str(i) for i in range(12)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  day Embed")
        self.embed_day = Embed(
            [str(i) for i in range(31)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  weekday Embed")
        self.embed_weekday = Embed(
            [str(i) for i in range(7)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  yearday Embed")
        self.embed_yearday = Embed(
            [str(i) for i in range(366)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  hour Embed")
        self.embed_hour = Embed(
            [str(i) for i in range(24)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  minute Embed")
        self.embed_minute = Embed(
            [str(i) for i in range(60)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  second Embed")
        self.embed_second = Embed(
            [str(i) for i in range(60)],
            embedding_size,
            representation="dense",
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        # Summed sizes of all of the embeddings.
        fc_layer_input_size = (
            self.year_fc.output_shape[0]
            + self.embed_month.output_shape[0]
            + self.embed_day.output_shape[0]
            + self.embed_weekday.output_shape[0]
            + self.embed_yearday.output_shape[0]
            + self.embed_hour.output_shape[0]
            + self.embed_minute.output_shape[0]
            + self.embed_second.output_shape[0]
            + 1  # for periodic_second_of_day.
        )

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=fc_layer_input_size,
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        :param inputs: The input vector fed into the encoder.
            Shape: [batch x DATE_INPUT_SIZE], type torch.int8
        :type inputs: Tensor
        """
        # ================ Embeddings ================
        input_vector = inputs.type(torch.int)

        scaled_year = self.year_fc(input_vector[:, 0:1].type(torch.float))
        embedded_month = self.embed_month(input_vector[:, 1:2] - 1)
        embedded_day = self.embed_day(input_vector[:, 2:3] - 1)
        embedded_weekday = self.embed_weekday(input_vector[:, 3:4])
        embedded_yearday = self.embed_yearday(input_vector[:, 4:5] - 1)
        embedded_hour = self.embed_hour(input_vector[:, 5:6])
        embedded_minute = self.embed_minute(input_vector[:, 6:7])
        embedded_second = self.embed_second(input_vector[:, 7:8])
        periodic_second_of_day = torch_utils.periodic(input_vector[:, 8:9].type(torch.float), 86400)

        hidden = torch.cat(
            [
                scaled_year,
                embedded_month,
                embedded_day,
                embedded_weekday,
                embedded_yearday,
                embedded_hour,
                embedded_minute,
                embedded_second,
                periodic_second_of_day,
            ],
            dim=1,
        )

        # ================ FC Stack ================
        # logger.debug('  flatten hidden: {0}'.format(hidden))

        hidden = self.fc_stack(hidden)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return DateEmbedConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([DATE_INPUT_SIZE])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape


@DeveloperAPI
@register_encoder("wave", DATE)
class DateWave(Encoder):
    def __init__(
        self,
        fc_layers: Optional[List[FCStack]] = None,
        num_fc_layers: int = 1,
        output_size: int = 10,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict] = None,
        activation: str = "relu",
        dropout: float = 0,
        encoder_config=None,
        **kwargs,
    ):
        """
        :param fc_layers: list of dictionaries containing the parameters of
            all the fully connected layers.
        :type fc_layers: List
        :param num_fc_layers: Number of stacked fully connected layers.
        :type num_fc_layers: Integer
        :param output_size: Size of each layer.
        :type output_size: Integer
        :param use_bias: bool determines where to use a bias vector.
        :type use_bias: bool
        :param weights_initializer: Initializer for the weights (aka kernel)
            matrix.
        :type weights_initializer: string
        :param bias_initializer: Initializer for the bias vector.
        :type bias_initializer: string
        :param norm: type of normalization to use 'batch' or 'layer'.
        :type norm: string, default None
        :param norm_params: parameters to pass to normalization function.
        :type norm_params: dictionary
        :param activation: Activation function to use.
        :type activation: string
        :param dropout: determines if there should be a dropout layer before
            returning the encoder output.
        :type dropout: float
        """
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        logger.debug("  year FCStack")
        self.year_fc = FCStack(
            first_layer_input_size=1,
            num_layers=1,
            default_output_size=1,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=None,
            default_norm_params=None,
            default_activation=None,
            default_dropout=dropout,
        )

        # Summed sizes of all of the embeddings.
        # Additional 8 for periodic_[month, day, ..., second_of_day].
        fc_layer_input_size = self.year_fc.output_shape[0] + 8

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=fc_layer_input_size,
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        :param inputs: The input vector fed into the encoder.
            Shape: [batch x DATE_INPUT_SIZE], type torch.int8
        :type inputs: Tensor
        """
        # ================ Embeddings ================
        input_vector = inputs.type(torch.float)
        scaled_year = self.year_fc(input_vector[:, 0:1])
        periodic_month = torch_utils.periodic(input_vector[:, 1:2], 12)
        periodic_day = torch_utils.periodic(input_vector[:, 2:3], 31)
        periodic_weekday = torch_utils.periodic(input_vector[:, 3:4], 7)
        periodic_yearday = torch_utils.periodic(input_vector[:, 4:5], 366)
        periodic_hour = torch_utils.periodic(input_vector[:, 5:6], 24)
        periodic_minute = torch_utils.periodic(input_vector[:, 6:7], 60)
        periodic_second = torch_utils.periodic(input_vector[:, 7:8], 60)
        periodic_second_of_day = torch_utils.periodic(input_vector[:, 8:9], 86400)

        hidden = torch.cat(
            [
                scaled_year,
                periodic_month,
                periodic_day,
                periodic_weekday,
                periodic_yearday,
                periodic_hour,
                periodic_minute,
                periodic_second,
                periodic_second_of_day,
            ],
            dim=1,
        )

        # ================ FC Stack ================
        # logger.debug('  flatten hidden: {0}'.format(hidden))

        hidden = self.fc_stack(hidden)
        return {"encoder_output": hidden}

    @staticmethod
    def get_schema_cls():
        return DateWaveConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([DATE_INPUT_SIZE])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape



# Implements https://github.com/ray-project/ray/pull/30598 ahead of Ray 2.2 release.

import math
from typing import Any, Callable, Dict, Optional, Type, TYPE_CHECKING, Union

import ray
from ray.air.config import RunConfig
from ray.tune.execution.trial_runner import _ResumeConfig
from ray.tune.impl.tuner_internal import TunerInternal
from ray.tune.trainable import Trainable
from ray.tune.tune_config import TuneConfig
from ray.tune.tuner import _SELF, _TUNER_INTERNAL, Tuner
from ray.tune.utils.node import _force_on_current_node

if TYPE_CHECKING:
    from ray.train.trainer import BaseTrainer


class TunerRay210(Tuner):
    """HACK(geoffrey): This is a temporary fix to support Ray 2.1.0.

    Specifically, this Tuner ensures that TunerInternalRay210 is called by the class.
    For more details, see TunerInternalRay210.
    """

    def __init__(
        self,
        trainable: Optional[
            Union[
                str,
                Callable,
                Type[Trainable],
                "BaseTrainer",
            ]
        ] = None,
        *,
        param_space: Optional[Dict[str, Any]] = None,
        tune_config: Optional[TuneConfig] = None,
        run_config: Optional[RunConfig] = None,
        # This is internal only arg.
        # Only for dogfooding purposes. We can slowly promote these args
        # to RunConfig or TuneConfig as needed.
        # TODO(xwjiang): Remove this later.
        _tuner_kwargs: Optional[Dict] = None,
        _tuner_internal: Optional[TunerInternal] = None,
    ):
        """Configure and construct a tune run."""
        kwargs = locals().copy()
        self._is_ray_client = ray.util.client.ray.is_connected()
        if _tuner_internal:
            if not self._is_ray_client:
                self._local_tuner = kwargs[_TUNER_INTERNAL]
            else:
                self._remote_tuner = kwargs[_TUNER_INTERNAL]
        else:
            kwargs.pop(_TUNER_INTERNAL, None)
            kwargs.pop(_SELF, None)
            if not self._is_ray_client:
                self._local_tuner = TunerInternalRay210(**kwargs)
            else:
                self._remote_tuner = _force_on_current_node(ray.remote(num_cpus=0)(TunerInternalRay210)).remote(
                    **kwargs
                )

    @classmethod
    def restore(
        cls,
        path: str,
        resume_unfinished: bool = True,
        resume_errored: bool = False,
        restart_errored: bool = False,
    ) -> "Tuner":
        """Restores Tuner after a previously failed run.

        All trials from the existing run will be added to the result table. The
        argument flags control how existing but unfinished or errored trials are
        resumed.

        Finished trials are always added to the overview table. They will not be
        resumed.

        Unfinished trials can be controlled with the ``resume_unfinished`` flag.
        If ``True`` (default), they will be continued. If ``False``, they will
        be added as terminated trials (even if they were only created and never
        trained).

        Errored trials can be controlled with the ``resume_errored`` and
        ``restart_errored`` flags. The former will resume errored trials from
        their latest checkpoints. The latter will restart errored trials from
        scratch and prevent loading their last checkpoints.

        Args:
            path: The path where the previous failed run is checkpointed.
                This information could be easily located near the end of the
                console output of previous run.
                Note: depending on whether ray client mode is used or not,
                this path may or may not exist on your local machine.
            resume_unfinished: If True, will continue to run unfinished trials.
            resume_errored: If True, will re-schedule errored trials and try to
                restore from their latest checkpoints.
            restart_errored: If True, will re-schedule errored trials but force
                restarting them from scratch (no checkpoint will be loaded).
        """
        resume_config = _ResumeConfig(
            resume_unfinished=resume_unfinished,
            resume_errored=resume_errored,
            restart_errored=restart_errored,
        )

        if not ray.util.client.ray.is_connected():
            tuner_internal = TunerInternalRay210(restore_path=path, resume_config=resume_config)
            return TunerRay210(_tuner_internal=tuner_internal)
        else:
            tuner_internal = _force_on_current_node(ray.remote(num_cpus=0)(TunerInternalRay210)).remote(
                restore_path=path, resume_config=resume_config
            )
            return TunerRay210(_tuner_internal=tuner_internal)


class TunerInternalRay210(TunerInternal):
    """HACK(geoffrey): This is a temporary fix to support Ray 2.1.0.

    This TunerInternal ensures that a division by zero is avoided when running zero-CPU hyperopt trials.
    This is fixed in ray>=2.2 (but not ray<=2.1) here: https://github.com/ray-project/ray/pull/30598
    """

    def _expected_utilization(self, cpus_per_trial, cpus_total):
        num_samples = self._tune_config.num_samples
        if num_samples < 0:  # TODO: simplify this in Tune
            num_samples = math.inf
        concurrent_trials = self._tune_config.max_concurrent_trials or 0
        if concurrent_trials < 1:  # TODO: simplify this in Tune
            concurrent_trials = math.inf

        actual_concurrency = min(
            (
                (cpus_total // cpus_per_trial) if cpus_per_trial else 0,
                num_samples,
                concurrent_trials,
            )
        )
        return (actual_concurrency * cpus_per_trial) / (cpus_total + 0.001)



import logging
from typing import Optional

import numpy as np
import torch
from torch.nn import BatchNorm1d, BatchNorm2d, LayerNorm, Module

from ludwig.utils.torch_utils import LudwigModule

logger = logging.getLogger(__name__)


# implementation adapted from https://github.com/dreamquark-ai/tabnet
class GhostBatchNormalization(LudwigModule):
    def __init__(
        self, num_features: int, momentum: float = 0.05, epsilon: float = 1e-3, virtual_batch_size: Optional[int] = 128
    ):
        super().__init__()
        self.num_features = num_features
        self.virtual_batch_size = virtual_batch_size
        self.bn = torch.nn.BatchNorm1d(num_features, momentum=momentum, eps=epsilon)

    def forward(self, inputs):
        batch_size = inputs.shape[0]

        if self.training and self.virtual_batch_size:
            splits = inputs.chunk(int(np.ceil(batch_size / self.virtual_batch_size)), 0)
            if batch_size == 1:
                logger.warning(
                    "Batch size is 1, but batch normalization requires batch size >= 2. Skipping batch normalization."
                    "Make sure to set `batch_size` to a value greater than 1."
                )
                # Skip batch normalization if the batch size is 1.
                return torch.cat(splits, 0)
            if batch_size % self.virtual_batch_size == 1:
                # Skip batch normalization for the last chunk if it is size 1.
                logger.warning(
                    f"Virtual batch size `{self.virtual_batch_size}` is not a factor of the batch size `{batch_size}`, "
                    "resulting in a chunk of size 1. Skipping batch normalization for the last chunk of size 1."
                )
            splits_with_bn = [self.bn(x) if x.shape[0] > 1 else x for x in splits]
            return torch.cat(splits_with_bn, 0)

        if batch_size != 1 or not self.training:
            return self.bn(inputs)
        return inputs

    @property
    def moving_mean(self) -> torch.Tensor:
        return self.bn.running_mean

    @property
    def moving_variance(self) -> torch.Tensor:
        return self.bn.running_var

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.num_features])

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.num_features])


norm_registry = {
    "batch_1d": BatchNorm1d,
    "batch_2d": BatchNorm2d,
    "layer": LayerNorm,
    "ghost": GhostBatchNormalization,
}


def create_norm_layer(norm: str, input_rank: int, num_features: int, **norm_params) -> Module:
    if norm == "batch":
        # We use a different batch norm depending on the input_rank.
        # TODO(travis): consider moving this behind a general BatchNorm interface to avoid this kludge.
        if input_rank not in {2, 3}:
            ValueError(f"`input_rank` parameter expected to be either 2 or 3, but found {input_rank}.")
        norm = f"{norm}_{input_rank-1}d"

    norm_cls = norm_registry.get(norm)
    if norm_cls is None:
        raise ValueError(
            f"Unsupported value for `norm` param: {norm}. Supported values are: {list(norm_registry.keys())}"
        )

    return norm_cls(num_features, **norm_params)
